///////////////////////////////////////////////////////////////////////////////
//
// Example Ericsson chatboard UART driver v0.03 beta!
// (c) 2003 Jouni 'Mr.Spiv/TRSi' Korhonen
//
//   http://www.deadcoderssociety.tk
//
// Note: this driver uses fully IRQ driven bidirectional
//       UART communication with the keyboard.
//       This code does not rely too much on the sdk.
//
//       The buffer handling etc in this driver is more or less
//       ad-hoc stuff typed in during one coding session so they might
//       contain some odd stuff :)
//
//  Versions:
//   v0.01 - Basic FIFO support & init for CHA-01.
//   v0.02 - Added init support for another breed of CHA-01 boards.
//   v0.03 - Added inactivity driver so that the chatboard won't start sleeping
//           after a long period of inactivity.
//
//
///////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <ctype.h>
#include "gp32.h"
#include "kbdrv.h"


#define KDB_DEBUG
#ifdef KDB_DEBUG
#include "gpstdio.h"
#include "gpdef.h"
#include "gpstdlib.h"
#include "gpgraphic.h"
#include "gpfont.h"
#include "gp32.h"

#endif


//
// There are two kinds of init sequences for CHA-01 (so far..)
//

static int modefix = 0;

//
//
//

static int sendACKZ( unsigned char * );
static void installRTCInt( void );
static void removeRTCInt( void );
static void setupRTCInt( void );

//
//
//

static unsigned char _kbdBuffer[KBDBUFSIZE];
static volatile int _kbdBufferLo;
static volatile int _kbdBufferHi;
static volatile int _kbdBufferCnt;

static unsigned char _kbdTemp[KBDTMPSIZE];
static int _kbdTempCnt;

static unsigned char _kbdOut[KBDTMPSIZE];
static volatile int _kbdOutCnt;
static volatile int _kbdOutLen;
static int _mode = sms_mode;
static int _idletime = 0;



//
// Some stuff to map AT-commands to key presses..
//
//
//

struct threeMap {
	int key;
	const char *map1;
	const char *map2;
};

struct twoMap {
	int key;
	const char *map;
};


static const struct threeMap cha01KeyMap[] = {
	{'1',	"1\",20\r",				"0c1111111111111111\"\r"},
	{'2',	"2\",20\r",				"1c222222222\"\r"},
	{'3',	"3\",20\r",				"1c333333\"\r"},
	{'4',	"4\",20\r",				"1c44444\"\r"},
	{'5',	"5\",20\r",				"1c5555\"\r"},
	{'6',	"6\",20\r",				"1c66666666\"\r"},
	{'7',	"7\",20\r",				"1c777777\"\r"},
	{'8',	"8\",20\r",				"1c888888\"\r"},
	{'9',	"9\",20\r",				"1c99999\"\r"},
	{'0',	"0\",20\r",				"1c0000000000000\"\r"},

	{'\r',	"eee\"\r",				"c\"\r"},			// BACKSPACE!


	{'!',	"1111\"\r",				"0c1111\"\r"},
	{'@',	"000\"\r",				"1c000\"\r"},		//
	{'#',	"#\"\r",				"1c#\"\r"},
	{'$',	"0000000\"\r",			"1c0000000\"\r"},
	{'%',	"000000\"\r",			"1c000000\"\r"},
	{'?',	"111\"\r",				"0c111\"\r"},
	{'&',	"00\"\r",				"1c00\"\r"},		//
	{'*',	"##\"\r",				"1c##\"\r"},
	{'(',	"11111111111111\"\r",	"0c11111111111111\"\r"},
	{')',	"111111111111111\"\r",	"0c111111111111111\"\r"},

	{'Q',	"77\"\r",				"1c77\"\r"},
	{'W',	"9\"\r",				"1c9\"\r"},
	{'E',	"33\"\r",				"1c33\"\r"},
	{'R',	"777\"\r",				"1c777\"\r"},
	{'T',	"8\"\r",				"1c8\"\r"},
	{'Y',	"999\"\r",				"1c999\"\r"},
	{'U',	"88\"\r",				"1c88\"\r"},
	{'I',	"444\"\r",				"1c444\"\r"},
	{'O',	"666\"\r",				"1c666\"\r"},
	{'P',	"7\"\r",				"1c7\"\r"},
	{'A',	"2\"\r",				"1c2\"\r"},
	{'S',	"7777\"\r",				"1c7777\"\r"},
	{'D',	"3\"\r",				"1c3\"\r"},
	{'F',	"333\"\r",				"1c333\"\r"},
	{'G',	"4\"\r",				"1c4\"\r"},
	{'H',	"44\"\r",				"1c44\"\r"},
	{'J',	"5\"\r",				"1c5\"\r"},
	{'K',	"55\"\r",				"1c55\"\r"},
	{'L',	"555\"\r",				"1c555\"\r"},
	{'Z',	"9999\"\r", 			"1c9999\"\r"},
	{'X',	"99\"\r",				"1c99\"\r"},
	{'C',	"222\"\r",				"1c222\"\r"},
	{'V',	"888\"\r",				"1c888\"\r"},
	{'B',	"22\"\r",				"1c22\"\r"},
	{'N',	"66\"\r",				"1c66\"\r"},
	{'M',	"6\"\r",				"1c6\"\r"},
	{'.',	"111111\"\r",			"0c111111\"\r"},
	{'a',	"<\"\r",				"<\"\r"}, // bottom arrow keys
	{'b',	">\"\r",				">\"\r"},
	{' ',	"1\"\r", 				"0c1\"\r"}, // spacebar
	{'[',	"2222222\"\r",			"1c2222222\"\r"}, // start of shift characters
	{']',	"22222222\"\r",			"1c22222222\"\r"},
	{'\\',	"3333\"\r",				"1c3333\"\r"},
	{' ',	"33333\"\r",			"1c3333\"\r"},
	{' ',	"4444\"\r",				"1c4444\"\r"},
	{':',	"1111111\"\r",			"0c1111111\"\r"},
	{' ',	"6666\"\r",				"1c6666\"\r"},
	{' ',	"6666666\"\r",			"1c6666666\"\r"},
	{' ',	"77777\"\r",			"1c77777\"\r"},
	{'n',	"8888\"\r",				"1c8888\"\r"}, // shift P - return
	{' ',	"88888\"\r",			"1c88888\"\r"}, // second line of shift
	{' ',	"2222\"\r",				"1c2222\"\r"},
	{'\'',	"1111111111\"\r",		"0c1111111111\"\r"},
	{' ',	"222222\"\r",			"1c222222\"\r"},
	{'"',	"111111111\"\r",		"0c111111111\"\r"},
	{' ',	"666666\"\r",			"1c666666\"\r"},
	{' ',	"00000000\"\r",			"1c00000000\"\r"},
	{' ',	"66666\"\r",			"1c66666\"\r"},
	{' ',	"22222\"\r",			"1c22222\"\r"},
	{'+',	"0\"\r",				"1c0\"\r"}, // third line of shift
	{'-',	"11\"\r",				"0c11\"\r"},
	{'_',	"00000000000000000\"\r","1c00000000000000000\"\r"},
	{'=',	"0000000000000000\"\r",	"1c0000000000000000\"\r"},
	{'/',	"0000\"\r",				"1c0000\"\r"},
	{'<',	"7777777\"\r",			"1c7777777\"\r"},
	{'>',	"77777777\"\r",			"1c77777777\"\r"},
	{',',	"11111\"\r",			"0c11111\"\r"},
	{'s',	"*\"\r",			"1c*\"\r"} // shift-space
};

#define CHA_KEY_NUMBER (sizeof(cha01KeyMap) / sizeof(struct threeMap))


static struct twoMap keyMapSort1[CHA_KEY_NUMBER];
static struct twoMap keyMapSort2[CHA_KEY_NUMBER];

//
//
//

static int cmp( const void *a, const void *b ) {
	return strcmp( ((const struct twoMap*)a)->map, ((const struct twoMap*)b)->map);
}

//
// few helper functions to get rid of the SDK
//

void installIRQ( int num, void (*irq)(void) ) {
	asm volatile(""
		"stmdb	sp!,{lr}	\n"
		"mov	r0,%0	\n"
		"mov	r1,%1	\n"
		"swi	#0x09		\n"
		"ldmia	sp!,{lr}"
		: 
		: "r"(num), "r"(irq)
		: "r0","r1");
}

//
//
//

static void removeIRQ( int num ) {
	asm volatile(""
		"stmdb	sp!,{lr}	\n"
		"mov	r0,%[_num]	\n"
		"swi	#0x0a		\n"
		"ldmia	sp!,{lr}"
		:
		: [_num] "r"(num)
		: "r0");
}

//
//
//

static void ARMDisableInt( void ) {
	asm volatile(""
		"mrs	r0,CPSR		\n"
		"orr	r0,r0,#192	\n"
		"msr	CPSR_fsxc,r0	\n"
		:
		:
		: "r0");
}

//
//
//

static void ARMEnableInt( void ) {
	asm volatile(""
		"mrs	r0,CPSR		\n"
		"bic	r0,r0,#128	\n"
		"msr	CPSR_fsxc,r0	\n"
		:
		:
		: "r0");
}

//
//
//
//
// UART0 Rx IRQ handler..
//
// More or less Ericsson specific IRQ handler.. as the
// keyboard only sends '\r' terminated ascii commands
// we just scan for them..
//
//

static void UART0RxInt(void) __attribute__ ((naked));
static void UART0RxInt(void) {
	asm volatile (
		"stmdb    r13!,{r0-r12,lr}"
	);
	//long *p = (long *)PALETTE;
	//static int foo = 0;

	unsigned char c;
	int cnt;
	
	//*p = foo+=4;


	if (rUFSTAT0 & 0x100) {
		//cnt += ((rUFSTAT0 >> 8) & 0x01);
		cnt = 16;
	} else {
		cnt = rUFSTAT0 & 0x0f;
	}
	





	while (cnt-- > 0) {
		c = rURXH0;

		_kbdTemp[_kbdTempCnt++] = c;

		if (c == 13) {
			int i;

			// Check if we still have space for a new input..

			if (_kbdBufferCnt < KBDMAXCNT) {
				_kbdBufferCnt++;

				for (i = 0; i < _kbdTempCnt; i++) {
					_kbdBuffer[_kbdBufferHi++] = _kbdTemp[i];
					_kbdBufferHi = _kbdBufferHi % KBDBUFSIZE;
				}
	
				// Send ACK

				_idletime = 0;
				sendACKZ("OK\r\n");
			}

			_kbdTempCnt = 0;
		}
	}

	asm volatile (
		"ldmia  r13!,{r0-r12,lr}        \n"
		"subs pc,lr,#4"
	);
}

//
// More or less Ericsson specific IRQ handler.. as the
// ACK string must be zero terminated "OK\r\n" this
// handler has been made to do so..
//
//

static void UART0TxInt(void) __attribute__ ((naked));
static void UART0TxInt(void) {
	asm volatile (
		"stmdb    r13!,{r0-r12,lr}"
	);

	while (_kbdOutCnt < _kbdOutLen) {
		if (((rUFSTAT0 >> 4) & 0x0f) <= 15) {
			rUTXH0 = _kbdOut[_kbdOutCnt++];
		} else {
			break;
		}
	}

	if (_kbdOutCnt == _kbdOutLen) {
		_kbdOutLen = 0;
		_kbdOutCnt = 0;
		//rUCON0 &= ~4;
	}

	asm volatile (
		"ldmia  r13!,{r0-r12,lr}        \n"
		"subs pc,lr,#4"
	);
}

//
// Note:
//  Called within an interrupt..

int sendACKZ( unsigned char *s ) {
#if 1
	while (_kbdOutLen > 0);
	_kbdOutLen = strlen(s) + 1;
	_kbdOutCnt = 0;
	memcpy(_kbdOut,s,_kbdOutLen);		// adds the '\0' also..

	// Start writing..
	rSRCPND = (1 << 28);
	rINTPND = (1 << 28);

	while (_kbdOutCnt < _kbdOutLen) {
		if (((rUFSTAT0 >> 4) & 0x0f) <= 15) {
			rUTXH0 = _kbdOut[_kbdOutCnt++];
		} else {
			break;
		}
	}
#else
	do {
		while ((rUTRSTAT0 & 0x2) == 0);
		rUTXH0 = *s++;
	} while (*s);
#endif

	return 0;
}


//
//
// Returns:
//   0 if completed, 1 if postInitKbd() needs to be called..
//
//

int preInitKbd(long baud, long pclk) {
	unsigned char buf[256];
	int n,i;

	//
	// Sort key mapping tables..
	//

	for (n = 0; n < CHA_KEY_NUMBER; n++) {
		keyMapSort1[n].key = cha01KeyMap[n].key;
		keyMapSort1[n].map = cha01KeyMap[n].map1;
		keyMapSort2[n].key = cha01KeyMap[n].key;
		keyMapSort2[n].map = cha01KeyMap[n].map2;
	}

	qsort(keyMapSort1,CHA_KEY_NUMBER,sizeof(struct twoMap),cmp);
	qsort(keyMapSort2,CHA_KEY_NUMBER,sizeof(struct twoMap),cmp);

	//
	//
	//
	// Set up UART0
	rCLKCON		|= 0x100;
	rULCON0		= 0x03;	// 8 bits, no-parity
	rUCON0		= 0x85;	// polled mode & Rx timeout irq
	rUFCON0		= 0x37;	// enable & reset FIFOs, Rx 16 bytes, Tx 0 bytes
	rUBRDIV0	= (unsigned int)(pclk / (baud * 16)) - 1;

	// set up variables..

	_kbdBufferLo  = 0;
	_kbdBufferHi  = 0;
	_kbdBufferCnt = 0;

	_kbdTempCnt   = 0;
	_kbdOutCnt    = 0;
	_kbdOutLen    = 0;
	_mode         = sms_mode;

	// set up IRQs

	installIRQ(23,UART0RxInt);	// Rx int
	installIRQ(28,UART0TxInt);	// Tx int

	//
	// Start the chatboard negotiation..
	//

	/*
	// -> AT<13>
	memset(buf,0,256);
	for (i=0;i<1000000;i++) {
	  getKbdRawNB(buf,256);
      	  if (strcmp(buf,"\0")) { break; }
	}

	// If the first received data was not AT<13> then the
	// keyboard is already intialized.. no further init is
	// required. Note that this string is lost..

	if (strcmp(buf,"AT\r")) { return 0; }

	// -> AT+CGMM<13>
	memset(buf,0,256);
	for (i=0;i<1000000;i++) {
	  getKbdRawNB(buf,256);
      	  if (strcmp(buf,"\0")) { break; }
	}
	if (strcmp(buf,"AT+CGMM\r")) { return -1; }

	// -> AT+CGMR<13>
	memset(buf,0,256);
	for (i=0;i<1000000;i++) {
	  getKbdRawNB(buf,256);
      	  if (strcmp(buf,"\0")) { break; }
	}
	if (strcmp(buf,"AT+CGMR\r")) { return -2; }
*/
	return 1;
}


//
//
//


int postInitKbd( void ) {
	unsigned char str[256];
	int y = 80;

	// -> AT*EDME?<13>
	getKbdRaw(str,256);
	if (strcmp(str,"AT*EDME?\r")) { return -1; }
	y += 10;

	// -> AT*ESVM?<13>
	getKbdRaw(str,256);
	if (strcmp(str,"AT*ESVM?\r")) { return -2; }
	y += 10;


	// -> AT+CKPD="eseee<see<<s>>ss1"<13> or AT*ESVM=0<13>
	getKbdRaw(str,256);
	if (strcmp(str,"AT+CKPD=\"eseee<see<<s>>ss1\"\r")) {
		if (strcmp(str,"AT*ESVM=0\r")) {
			return -3;
		}
		modefix = 1;
	}
	y += 10;

	if (modefix) {
		getKbdRaw(str,256);
		if (strcmp(str,"AT+CKPD=\"eseee<see<<s>>ss1\"\r")) { return -5; }
		y += 10;
	}

	// -> AT+CKPD="c",25<13>
	getKbdRaw(str,256);
	if (strcmp(str,"AT+CKPD=\"c\",25\r")) { return -4; }
	y += 10;

	if (modefix) {
		getKbdRaw(str,256);
		if (strcmp(str,"AT*ESVM=0\r")) { return -6; }
	}

	return 0;
}


//
//
//

void removeKbd( void ) {
	removeIRQ(23);	// Rx int
	removeIRQ(28);	// Tx int
}


//
//
//

int resolveKeyMapping( unsigned char *s ) {
	static int state = state_none;
	struct twoMap bcmp;
	struct twoMap *ptr;
	int n;

	switch (state) {
	case state_edme:
			  sprintf(debug,"edme*%s*",s);
		if (!strcmp(s,"AT*ESVM?\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_esvm;
			return 0;
		} else {
			state = state_none;
			return -1;
		}
		break;
	case state_esvm:
			  sprintf(debug,"esvm*%s*",s);
		if (!strcmp(s,"AT+CKPD=\"eseee<see<<s>>ss1\"\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_see;
			return 0;
		} else if (!strcmp(s,"AT+CKPD=\"eseee<see<<s>>ss\"\r")) {
			// Phonebook pressed
			state = state_ckpd;
			return 0;
		} else if (!strcmp(s,"AT*ESVM=0\r")) {
			state = state_esvm0;
			return 0;
		} else {
			state = state_none;
			return -1;
		}
		break;

	case state_see:
			  sprintf(debug,"see*%s*",s);
		if (!strcmp(s,"AT+CKPD=\"c\",25\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_ckpd;
			return 0;
		} else {
			state = state_none;
			return -1;
		}
		break;

	case state_esvm0:
			  sprintf(debug,"esvm0*%s*",s);
		if (!strcmp(s,"AT+CKPD=\"eseee<see<<s>>ss1\"\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_see0;
			return 0;
		} else {
			state = state_none;
			return -1;
		}
		break;

	case state_see0:
			  sprintf(debug,"see0*%s*",s);
		if (!strcmp(s,"AT+CKPD=\"c\",25\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_c;
			return 0;
		} else {
			state = state_none;
			return -1;
		}
		break;

	case state_c:
			  sprintf(debug,"c*%s*",s);
		if (!strcmp(s,"AT*ESVM=0\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_c;
			return 0;
		} else {
			state = state_none;
			return -1;
		}
		break;

	case state_none: case state_ckpd:
			  sprintf(debug,"ckpd/none*%s*",s);
		if (!strcmp(s,"AT*EDME?\r")) {
			// SMS pressed.. probably due inactivity etc..
			state = state_edme;
			return 0;
		}
		if (!strncmp(s,"AT+CKPD=\"",9)) {
			// Now we know it is a normal key press..
			// skip start of the AT-command..
			s += 9;
			state = state_ckpd;

			// Do a binary search.. keymap1

			bcmp.key = -1;
			bcmp.map = s;

			if ((ptr = bsearch(&bcmp,keyMapSort1,CHA_KEY_NUMBER,sizeof(struct twoMap),cmp))) {
			  //			  sprintf(debug,"*%s*",s);
				return ptr->key;
			} else if ((ptr = bsearch(&bcmp,keyMapSort2,CHA_KEY_NUMBER,sizeof(struct twoMap),cmp))) {
			  //			  sprintf(debug,"*%s*",s);
				return ptr->key;
			} else {
				return -1;
			}
		} else {
			// unknown state..
			state = state_none;
			return -1;
		}
		break;

	default:
		break;
	}

	// some special key state..

	// AT+CKPD="",20  -> num
	// AT+CKPD=""     -> char

	// www-key
	// AT*EDME?
	// AT*ESVM?
	// AT+CKPD="eseee<see<<s>>ss1"
	// AT+CKPD="c",25
	// AT+CKPD="91c91c911111110c"

	// Attachment key
	// AT+CKPD="#11111110c"

	// email-key
	// AT*EDME?
	// AT*ESVM?
	// AT+CKPD="eseee<see<<s>>ss1"
	// AT+CKPD="c",25
	// AT+CKPD="8666*1111111*0c"

	// sms-key
	// AT*EDME?
	// AT*ESVM?
	// AT+CKPD="eseee<see<<s>>ss1"
	// AT+CKPD="c",25

	// phonebool-key
	// AT*EDME?
	// AT*ESVM?
	// AT+CKPD="eseee<see<<s>>ss"

	return -1;
}



//
// Note:
//  This call blocks.
//

int getKbdRaw( unsigned char *buf, int len ) {
	unsigned char c;
	int n = 0;

	while (_kbdBufferLo == _kbdBufferHi);
	ARMDisableInt();

	while (n < len) {
		buf[n++] = c = _kbdBuffer[_kbdBufferLo++];
		_kbdBufferLo = _kbdBufferLo % KBDBUFSIZE;
		if (c == 13) { 
			buf[n] = '\0';
			break;
		}
	}
	_kbdBufferCnt--;

	ARMEnableInt();
	return n;
}

//
// Note:
//  This call is non-blocking.
//

int getKbdRawNB( unsigned char *buf, int len ) {
	unsigned char c;
	int n = 0;

	if (_kbdBufferLo == _kbdBufferHi) {
		return 0;
	}

	ARMDisableInt();

	while (n < len) {
		buf[n++] = c = _kbdBuffer[_kbdBufferLo++];
		_kbdBufferLo = _kbdBufferLo % KBDBUFSIZE;
		if (c == 13) { 
			buf[n] = '\0';
			break;
		}
	}
	_kbdBufferCnt--;

	ARMEnableInt();
	return n;
}
